directory # unused attribute (mech_client/acn.py:85)
directory # unused attribute (mech_client/acn.py:100)
directory # unused attribute (mech_client/acn.py:116)
payment_data # unused variable (mech_client/interact.py:111)
AGENT_QUERY_TEMPLATE # unused variable (mech_client/subgraph.py:30)